import random
from gametheorylab.axelrod_interactive.strategy import Strategy

class GrimAttacker(Strategy):
    """Class that implements the Grim Attacker strategy. See the provided documentation for details.

    Args:
        defect_streak (int): The number of times opponent defects in a row before conclusively being identified as the grim trigger. Defaults to 3.
        coop_streak (int): The number of times for which the strategy initially cooperates before defecting to test for Grim Trigger's behavior. Defaults to 2.
    """
    def __init__(self, defect_streak: int=3, coop_streak: int=2):
        # Validation
        if not isinstance(defect_streak, int):
            raise ValueError(f"Parameter 'defect_streak' must be an integer! (You entered {defect_streak})")
        if not isinstance(coop_streak, int):
            raise ValueError(f"Parameter 'coop_streak' must be an integer! (You entered {coop_streak})")

        super().__init__()
        self.opp_is_grim = None

        self.base_cs = coop_streak
        self.coop_streak = coop_streak

        self.base_ds = defect_streak
        self.defect_streak = self.base_ds

        self.prob_of_cooperation = 0.15

        self.name = "Grim Attacker"
        self.__doc__ = """This strategy has one mission: to beat Grim Trigger in the round robin tournament. The 
        strategy starts off by cooperating 'coop_streak' times, then defecting once and then cooperating again 
        'defect_streak' times, while observing the opponent's behavior.
        
        If the opponent defects throughout the testing period, the strategy concludes that the opponent is the Grim 
        Trigger, and then defects for the remainder of the round. Otherwise, and if he defects unexpectedly (such as 
        while no defects have been made), the strategy concludes he is NOT the Grim Trigger, and he proceeds to 
        mirror GTFT's strategy with a 0.15 chance of forgiveness."""

    def additional_prep(self):
        self.defect_streak = self.base_ds
        self.coop_streak = self.base_cs
        self.opp_is_grim = None

    def update_history(self, opp_history, self_history):
        if self.opp_is_grim is not None: return

        if opp_history:
            if not opp_history[-1]:
                if self.coop_streak > -1:
                    self.opp_is_grim = False
                elif self.defect_streak == 0 and self.opp_is_grim is None:
                    self.opp_is_grim = True
            else:
                if not self_history[-1]:
                    return
                if self.coop_streak == -1 and self.defect_streak > 0:
                    self.opp_is_grim = False



    def move(self, opp_history, self_history, opp_score, self_score):
        if self.opp_is_grim is None:
            while self.coop_streak > 0:
                self.coop_streak -= 1
                return True

            if self.coop_streak == 0:
                self.coop_streak -= 1
                return False

            while self.defect_streak > 0:
                self.defect_streak -= 1
                return True

        else:
            if self.opp_is_grim:
                return False

            elif not opp_history[-1]:
                return random.random() < self.prob_of_cooperation

            return True



from .settings import PAYOFFS

payoff_dict = PAYOFFS.copy()
def set_payoffs(t=5, r=3, p=1, s=0):
    """Sets the payoff matrix for the session. It MUST be called immediately after loading the library in, otherwise a payoff matrix won't exist."""
    payoff_dict["Temptation"] = t
    payoff_dict["Reward"] = r
    payoff_dict["Punishment"] = p
    payoff_dict["Sucker"] = s

def get_payoffs():
    """Returns the payoff matrix for the session."""
    return payoff_dict



from gametheorylab.axelrod_interactive.strategy import Strategy

class Gradual(Strategy):
    """Class that implements the Gradual strategy. See the provided documentation for details."""
    def __init__(self):
        super().__init__()
        self.punishment = 0
        self.anger = 0

        self.name = "Gradual"
        self.__doc__ = ("This strategy works by punishing every defection with an increasing sequence of defections, "
                        "then returning to cooperation.")

    def additional_prep(self):
        self.punishment = 0
        self.anger = 0

    def update_history(self, opp_history, self_history):
        if opp_history:
            if not opp_history[-1] and self.punishment == 0:
                self.anger += 1
                self.punishment = self.anger

    def move(self, opp_history, self_history, opp_score, self_score):
        if self.punishment > 0:
            self.punishment -= 1
            return False
        return True





from gametheorylab.axelrod_interactive.strategy import Strategy

class SuspiciousTitForTat(Strategy):
    """Class that implements the Suspicious Tit for Tat strategy. See the provided documentation for details."""
    def __init__(self):
        super().__init__()
        self.name = "Suspicious Tit for Tat"
        self.__doc__ = ("Similar with Tit for Tat. However, this strategies starts off by defecting, and then just "
                        "plays the opponents last move.")

    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history:
            return False
        return opp_history[-1]


import os
import importlib
import inspect
import pkgutil
from gametheorylab.axelrod_interactive.strategy import Strategy

STRATEGIES = {}

strategy_dir = os.path.dirname(__file__)

for finder, module_name, is_pkg in pkgutil.iter_modules([strategy_dir]):
    try:
        module = importlib.import_module(f".{module_name}", package=__package__)
    except Exception as e:
        print(f"Warning: Failed to import {module_name}: {e}")
        continue

    for name, obj in inspect.getmembers(module, inspect.isclass):
        if issubclass(obj, Strategy) and obj is not Strategy and name != "User":
            globals()[name] = obj
            STRATEGIES[name] = obj

__all__ = list(STRATEGIES.keys())

locals().update(STRATEGIES)

import sys

def __getattr__(name):
    if name in STRATEGIES:
        return STRATEGIES[name]
    raise AttributeError(f"module {__name__} has no attribute {name}")

sys.modules[__name__].__dict__.update(STRATEGIES)





from gametheorylab.axelrod_interactive.strategy import Strategy

class AlwaysDefect(Strategy):
    """Class that implements the Always Defect strategy."""
    def __init__(self):
        super().__init__()
        self.name = "Always Defect"
        self.__doc__ = "This is a simple strategy that starts by defecting, and defects for the entire round."

    def move(self, opp_history, self_history, opp_score, self_score):
        return False


import random
from gametheorylab.axelrod_interactive.strategy import Strategy

class CoinFlip(Strategy):
    """Class that implements the 'coin flip' (aka random) strategy.

    Args:
        prob_of_cooperation (float): The probability that the strategy cooperates for any given round. Defaults to 0.5.
    """
    def __init__(self, prob_of_cooperation=0.5):
        # Validation
        if not 0 < prob_of_cooperation < 1:
            if prob_of_cooperation <= 0:
                string = f"({prob_of_cooperation} <= 0)"
            else:
                string = f"({prob_of_cooperation} >= 1)"
            raise ValueError(f"Parameter 'prob_of_cooperation' is out of range! " + string)

        super().__init__()
        self.prob_of_cooperation = prob_of_cooperation

        self.name = f"Coin Flip (p = {self.prob_of_cooperation})"
        self.__doc__ = "This strategy randomly cooperates or defects at any given moment in the game."

    def move(self, opp_history, self_history, opp_score, self_score):
        return random.random() < self.prob_of_cooperation



import random
from gametheorylab.axelrod_interactive.strategy import Strategy

class Joss(Strategy):
    """Class that implements the Joss strategy. See provided documentation for details.
    Args:
        prob_of_defection (float): Probability of defecting when the opponent cooperates. Defaults to 0.1. Note that it must be in (0, 1).
    """
    def __init__(self, prob_of_defection=0.1):
        # Validation
        if not 0 < prob_of_defection < 1:
            if prob_of_defection <= 0:
                string = f"({prob_of_defection} <= 0)"
            else:
                string = f"({prob_of_defection} >= 1)"
            raise ValueError(f"Parameter 'prob_of_defection' is not in range! " + string)

        super().__init__()
        self.prob_of_defection = prob_of_defection

        self.name = f"Joss (p = {self.prob_of_defection})"
        self.__doc__ = ("This strategy plays like Tit for Tat, except when the opponent cooperates, it sometimes "
                        "defects (precisely, with probability 'prob_of_defection').")

    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history:
            return True

        if not opp_history[-1]:
            return False

        return random.random() >= self.prob_of_defection



from gametheorylab.axelrod_interactive.strategy import Strategy

class TitForKTats(Strategy):
    """Class that implements the Tit for 'K' Tats strategy. See the provided documentation for details.
    Args:
        k (int): Number of defects by opponent that warrants a defect from the strategy (number of tats for one tit). Defaults to 2.
    """
    def __init__(self, k: int=2):
        # Validation
        if not isinstance(k, int):
            raise ValueError(f"Can you defect a fractional number of times? (You entered k = {k}.)")
        super().__init__()
        self.k = k
        self.counter = self.k
        self.name = f"Tit for {self.k} tats"
        self.__doc__ = (f"This strategy is similar to Tit for Tat, with the slight variation that a defect (tit) is "
                        f"played after {self.k} defects in a row ({self.k} tats).")

    def additional_prep(self):
        self.counter = self.k

    def update_history(self, opp_history, self_history):
        if opp_history:
            if not opp_history[-1] and self.counter > 0:
                self.counter -= 1
            else:
                self.counter = self.k

    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history or self.counter > 0:
            return True

        self.counter = self.k
        return False


from gametheorylab.axelrod_interactive.strategy import Strategy

class SoftMajority(Strategy):
    """Class that implements the Soft Majority strategy."""
    def __init__(self):
        super().__init__()
        self.name = "Soft Majority"
        self.__doc__ = ("This strategy starts by cooperating, and cooperates if the opponent cooperates at least as "
                        "much as it defects, and defects otherwise.")

    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history:
            return True
        return self.opps_coop_count >= self.opps_defect_count



from gametheorylab.axelrod_interactive.strategy import Strategy

class Pavlov(Strategy):
    """Class that implements the Pavlov strategy. See the provided documentation for details."""
    def __init__(self):
        super().__init__()
        self.name = "Pavlov"
        self.__doc__ = ("Also known as 'Win-Stay-Lose-Shift', this strategy starts by cooperating, and from then on, "
                        "he repeats his last play if it won him the last round, and changes play otherwise.")

    def move(self, opp_history, self_history, opp_score, self_score):
        # Neat, isn't it?
        return not opp_history or opp_history[-1]


import pandas as pd

class Result:
    def __init__(self, data):
        self.data = data

    def to_csv(self, file_name: str):
        df = pd.DataFrame(self.data)
        df.to_csv(file_name, index=False)


from gametheorylab.axelrod_interactive.strategy import Strategy

class ReverseTitForTat(Strategy):
    """Class that implements the Reverse Tit for Tat strategy."""
    def __init__(self):
        super().__init__()
        self.name = "Reverse Tit for Tat"
        self.__doc__ = ("Similar with Tit for Tat. However, this strategies starts off by defecting, and then plays "
                        "the reverse of the opponents' last move (so if the opponent defects, it cooperates, "
                        "and if the opponent cooperates, it defects).")

    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history:
            return False
        return not opp_history[-1]


import random
from gametheorylab.axelrod_interactive.strategy import Strategy

class AdaptiveGTFT(Strategy):
    """Class that implements the Adaptive GTFT (Generous Tit for Tat) strategy. See provided documentation for details.

    Args:
        base_generosity (float): Initial probability of forgiving a defection. Must be in (0, 1).
        delta_p (float): Parametrizes how sensitive the strategy's 'generosity' is to perceived cooperation/defection. Must also be in (0, 1).
    """
    def __init__(self, base_generosity: float=0.5, delta_p: float=0.1):
        # Validation
        if not 0 < base_generosity < 1:
            if base_generosity <= 0:
                string = f"({base_generosity} <= 0)"
            else:
                string = f"({base_generosity} >= 1)"
            raise ValueError("Parameter 'base_generosity' is out of range! " + string)
        if not 0 < delta_p < 1:
            if delta_p <= 0:
                string = f"({delta_p} <= 0)"
            else:
                string = f"({delta_p} >= 1)"
            raise ValueError("Parameter 'delta_p' is out of range! " + string)

        super().__init__()
        self.delta_p = delta_p
        self.base_generosity = base_generosity
        self.prob_of_cooperation = base_generosity

        self.name = f"Adaptive G-TFT (p = {self.prob_of_cooperation:.3f}, dp = {self.delta_p:.3f})"
        self.__doc__ = ("This strategy is just like Generous Tit for Tat, but the chances of it forgiving future "
                        "defections are increased with each cooperation (and decreased with each defection), "
                        "as parameterized by 'delta_p'. Note that both 'base_generosity' and 'delta_p' must be "
                        "between 0 and 1.")

    def additional_prep(self):
        self.prob_of_cooperation = self.base_generosity

    def update_history(self, opp_history, self_history):
        if opp_history:
            if opp_history[-1]:
                self.prob_of_cooperation += (1 - self.prob_of_cooperation) * self.delta_p
            else:
                self.prob_of_cooperation -= self.prob_of_cooperation * self.delta_p

    def move(self, opp_history, self_history, opp_score, self_score):
        if opp_history and not opp_history[-1]:
            return random.random() >= self.prob_of_cooperation
        return True




from gametheorylab.axelrod_interactive.strategy import Strategy

class Tranquilizer(Strategy):
    """Class that implements the Tranquilizer strategy. See the provided documentation for details.

    Args:
        period (int): The length of the strategy's initial cooperation period. Defaults to 5.
    """
    def __init__(self, period=5):
        # Validation
        if not isinstance(period, int):
            raise ValueError(f"Can you cooperate a fractional number of times? (You entered period = {period}).")

        super().__init__()
        self.period = period
        self.counter = 0

        self.name = f"Tranquilizer (period = {self.period})"
        self.__doc__ = ("One of the sneakiest strategies yet, the tranquilizer starts off by cooperating, then defects "
                        "periodically to exploit kindness. If it perceives more-than-normal defection, "
                        "the cooperation period is reduced until it reaches 2.")

    def update_history(self, opp_history, self_history):
        if not opp_history: return

        if not opp_history[-1]:
            self.opps_defect_count += 1
        if self.counter == 0:
            if self.opps_defect_count > 1 + int(not opp_history[-1]):
                if self.period > 1:
                    self.period -= 1
            self.opps_defect_count = 0

    def move(self, opp_history, self_history, opp_score, self_score):
        if self.counter < self.period:
            self.counter += 1
            return True

        self.counter = 0
        return False



from abc import ABC, abstractmethod

class Strategy(ABC):
    """The base class that is inherited by all concrete strategy classes."""
    name: str
    prob_of_cooperation: float
    prob_of_defection: float
    opps_defect_count: int
    self_defect_count: int
    opps_coop_count: int
    self_coop_count: int

    def __init__(self):
        pass

    def __repr__(self):
        return self.name

    def additional_prep(self):
        pass

    def prep_for_match(self):
        self.opps_defect_count = 0
        self.self_defect_count = 0
        self.opps_coop_count = 0
        self.self_coop_count = 0
        self.additional_prep()

    def update_history(self, opp_history, self_history):
        if self_history:
            if self_history[-1]:
                self.self_coop_count += 1
            else:
                self.self_defect_count += 1

            if opp_history[-1]:
                self.opps_coop_count += 1
            else:
                self.opps_defect_count += 1

    def play(self, opp_history, self_history, opp_score, self_score):
        self.update_history(opp_history, self_history)
        return self.move(opp_history, self_history, opp_score, self_score)

    @abstractmethod
    def move(self, opp_history, self_history, opp_score, self_score):
        pass





import random
from gametheorylab.axelrod_interactive.strategy import Strategy

class GenerousTitForTat(Strategy):
    """Class that implements the Generous Tit for Tat strategy. See the provided documentation for details.

    Args:
        chance_of_forgiveness (float): Literally the chance that the strategy forgives a defection. Note that 'chance_of_forgiveness' must be in (0, 1).
    """
    def __init__(self, chance_of_forgiveness=0.1):
        if not 0 < chance_of_forgiveness < 1:
            if chance_of_forgiveness <= 0:
                string = f"({chance_of_forgiveness} <= 0)"
            else:
                string = f"({chance_of_forgiveness} >= 1)"
            raise ValueError(f"Parameter 'chance_of_forgiveness' is out of range! " + string)
        super().__init__()
        self.chance_of_forgiveness = chance_of_forgiveness
        self.name = f"Generous Tit for Tat (p = {chance_of_forgiveness})"
        self.__doc__ = ("This is a more generous version of Tit for Tat which probabilistically forgives a defection, "
                        "as parameterized by 'chance_of_forgiveness'. Note that 'chance_of_forgiveness' must hence be"
                        " in (0, 1).")
    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history:
            return True
        if not opp_history[-1]:
            return random.random() < self.chance_of_forgiveness
        return True





from gametheorylab.axelrod_interactive.strategy import Strategy
from gametheorylab.axelrod_interactive import get_payoffs
from gametheorylab.axelrod_interactive.result import Result

class Arena:
    """Class that simulates a head-to-head match between two strategies.

    Args:
        s1 (Strategy): One of the strategies playing the match.
        s2 (Strategy): The other strategies playing the match.
        num_rounds (int): The number of rounds in the match. Defaults to 10.
        show_results: Flag that determines whether the results of the match should be printed onto the console. Defaults to False.
    """
    payoffs = get_payoffs()
    def __init__(self, s1: Strategy, s2: Strategy, num_rounds: int=10, show_results: bool=True):
        # Validation
        if not isinstance(s1, Strategy):
            raise ValueError("Parameter 's1' is not a Strategy.")
        if not isinstance(s2, Strategy):
            raise ValueError("Parameter 's2' is not a Strategy.")
        if not isinstance(num_rounds, int):
            raise ValueError(f"Parameter 'num_rounds' must be an integer! (You entered {num_rounds}).")
        if not isinstance(show_results, bool):
            raise ValueError(f"Parameter 'show_results' must be a boolean! (You entered {show_results}).")

        self.show_results = show_results
        self.s1 = s1
        self.s2 = s2
        self.num_rounds = num_rounds
        self.s1_history = []
        self.s2_history = []
        self.s1_score = 0
        self.s2_score = 0

    def run(self):
        choice1 = self.s1.play(self.s2_history, self.s1_history, self.s2_score, self.s1_score)
        choice2 = self.s2.play(self.s1_history, self.s2_history, self.s1_score, self.s2_score)

        self.s1_history.append(choice1)
        self.s2_history.append(choice2)

        if choice1 and choice2:
            self.s1_score += self.payoffs["Reward"]
            self.s2_score += self.payoffs["Reward"]

        elif choice1:
            self.s1_score += self.payoffs["Sucker"]
            self.s2_score += self.payoffs["Temptation"]

        elif choice2:
            self.s1_score += self.payoffs["Temptation"]
            self.s2_score += self.payoffs["Sucker"]

        else:
            self.s1_score += self.payoffs["Punishment"]
            self.s2_score += self.payoffs["Punishment"]

    def play_round(self):
        """This simulates the two players going at it for the specified number of times. Who's going to win this round?"""
        self.s1.prep_for_match()
        self.s2.prep_for_match()
        for _ in range(self.num_rounds):
            if self.s1_score < 0 or self.s2_score < 0:
                break
            self.run()

        if self.show_results:
            print(f"{self.s1}'s moves: ")
            print(["C" if move else "D" for move in self.s1_history])

            print(f"{self.s2}'s moves: ")
            print(["C" if move else "D" for move in self.s2_history])

            print(f"Results: {self.s1_score} - {self.s2_score}")
            print("\n")

        result_dict = {"Move": list(range(1, self.num_rounds + 1)) + ["SCORE"],
                       self.s1.name: ["C" if move else "D" for move in self.s1_history] + [self.s1_score],
                       self.s2.name: ["C" if move else "D" for move in self.s2_history] + [self.s2_score]}
        return Result(result_dict)



from gametheorylab.axelrod_interactive.strategy import Strategy

class TitForTat(Strategy):
    """Class that implements the Tit for Tat strategy."""
    def __init__(self):
        super().__init__()
        self.name = "Tit for Tat"
        self.__doc__ = "This strategies starts off by cooperating, and then just plays the opponents last move."

    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history:
            return True
        return opp_history[-1]


from gametheorylab.axelrod_interactive.strategy import Strategy

class PavlovVariant(Strategy):
    """Class that implements the Pavlov (Variant) strategy. See the provided documentation for details."""
    def __init__(self):
        super().__init__()
        self.name = "Pavlov (Variant)"
        self.__doc__ = ("Similar to Pavlov, Pavlov (Variant) switches or stays at a move based on his position in the "
                        "game. This time, Pavlov (variant) switches exactly when he's losing, and stays otherwise.")

    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history:
            return True
        return (not self_history[-1]) == self_score < opp_score





from gametheorylab.axelrod_interactive.strategy import Strategy

class SoftGrim(Strategy):
    """Class that implements the Soft Grim strategy. See the provided documentation for details.

    Args:
        beef_period (int): The period during which the strategy defects as a result of the opponent's prior defection.
    """
    def __init__(self, beef_period: int=100):
        # Validation
        if not isinstance(beef_period, int):
            raise ValueError(f"Can you defect a fractional number of times? (You entered {beef_period}.)")
        super().__init__()
        self.beef_period = beef_period
        self.name = f"Soft Grim (period = {self.beef_period})"
        self.__doc__ = ("This strategy is like Grim Trigger, but it cooperates after a long period, as specified by "
                        "'beef_period'.")
        self.beef = 0

    def additional_prep(self):
        self.beef = 0

    def update_history(self, opp_history, self_history):
        if opp_history and not opp_history[-1]:
            if self.beef == 0 and self_history[-1]: self.beef = self.beef_period

    def move(self, opp_history, self_history, opp_score, self_score):
        if self.beef > 0:
            self.beef -= 1
            return False
        return True





from gametheorylab.axelrod_interactive.result import Result
from gametheorylab.axelrod_interactive.strategy import Strategy
from gametheorylab.axelrod_interactive.arena import Arena

class Tournament:
    def __init__(self, strategies: list[Strategy], mode: str="round robin", num_repeats: int=5, num_rounds: int=200):
        """Simulates a tournament with the given strategies.

           Args:
              strategies (list[Strategy]): the strategies to compete in the tournament. Number of strategies must be at least 3.
              mode (str): the mode of tournament to play. Currently supports three modes: 'round robin', 'four ways' and 'round of 16'. If 'round robin' is chosen, any number of strategies can be passed in. If 'four ways' is chosen, the number of strategies passed in must be exactly 4. If 'round of 16' is chosen, the number of strategies must be exactly 16. The default mode is 'round robin'.
              num_repeats (int): Number of times a single tournament is repeated. Defaults to 5.
              num_rounds (int): Number of rounds per match between any two strategies. Defaults to 200.
           """
        # Validation
        if len(strategies) < 3:
            raise ValueError(f"Strategies list is too small! ({len(strategies)} < 3)")
        if mode not in {"round of 16", "round robin", "four ways"}:
            raise ValueError(f"Mode {mode} either doesn't exist or is not supported.")
        if mode == "four ways" and len(strategies) != 4:
            raise ValueError(f"Mode {mode} requires exactly 4 strategies (but you passed in {len(strategies)}).")
        if mode == "round of 16" and len(strategies) != 16:
            raise ValueError(f"Mode {mode} requires exactly 16 strategies (but you passed in {len(strategies)}).")
        if not isinstance(num_rounds, int):
            raise ValueError(f"Parameter 'num_rounds' must be an integer! (You entered {num_rounds}.)")
        if not isinstance(num_repeats, int):
            raise ValueError(f"Parameter 'num_repeats' must be an integer! (You entered {num_repeats}.)")

        self.base_strategies = strategies
        self.strategies = strategies

        self.round_robin_results = {}
        self.num_players = len(self.strategies)
        self.mode = mode

        self.base_num_repeats = num_repeats
        self.num_repeats = self.base_num_repeats

        self.base_num_rounds = num_rounds
        self.num_rounds = self.base_num_rounds

    def __reset(self):
        self.strategies = self.base_strategies
        self.round_robin_results = {}
        self.num_players = len(self.strategies)
        self.num_repeats = self.base_num_repeats
        self.num_rounds = self.base_num_rounds

    def show_info(self):
        string = ""
        if self.mode == "round robin":
            string = f"Round robin ({self.num_rounds} rounds per match, {self.num_repeats} repetitions)"
        elif self.mode == "round of 16":
            string = "Round of 16"
        elif self.mode == "four ways":
            string = "Four ways"

        print("Mode:", string)
        print("Players:")
        for i in range(len(self.strategies)):
            print(f"{i + 1}:", self.strategies[i].name)

    def __play_round_robin(self, show_results=True):
        for _ in range(self.num_repeats):
            for i in range(self.num_players):
                for j in range(i, self.num_players):
                    player1 = self.strategies[i]
                    player2 = self.strategies[j]
                    arena = Arena(player1, player2, num_rounds=self.num_rounds, show_results=False)
                    result_dict = arena.play_round().data
                    p1_score, p2_score = result_dict[player1.name][-1], result_dict[player2.name][-1]
                    self.round_robin_results[player1.name] = self.round_robin_results.get(player1.name, 0) + p1_score
                    self.round_robin_results[player2.name] = self.round_robin_results.get(player2.name, 0) + p2_score
        for player in self.strategies:
            self.round_robin_results[player.name] /= (self.num_players * self.num_repeats)

        self.strategies.sort(key=lambda x: self.round_robin_results[x.name], reverse=True)
        winner = self.strategies[0].name

        if show_results:
            print(f"The winner is: {winner}!")
            print("Results: \n")

            for s in self.strategies:
                print(f"{s.name}: {round(self.round_robin_results[s.name], 2)}")

        leaderboard = {(name := self.strategies[i].name): (i + 1, self.round_robin_results[name]) for i in range(self.num_players)}
        results_dict = {
            "Rank": [],
            "Player": [],
            "Score": []
        }
        for player in leaderboard:
            results_dict["Rank"].append(leaderboard[player][0])
            results_dict["Player"].append(player)
            results_dict["Score"].append(leaderboard[player][1])
        return Result(results_dict)

    def __play_four_ways(self, show_results=True):
        if show_results:
            print("--------------------------------SEMI-FINALS----------------------------------")
        string = "Players: "
        for strategy in self.strategies:
            string += strategy.name + ", "
        print(string.strip(", ") + "\n")

        for i in range(4):
            for j in range(i + 1, 4):
                player1 = self.strategies[i]
                player2 = self.strategies[j]
                arena = Arena(player1, player2, num_rounds=self.num_rounds, show_results=False)
                result_dict = arena.play_round().data
                p1_score, p2_score = result_dict[player1.name][-1], result_dict[player2.name][-1]
                self.round_robin_results[player1.name] = self.round_robin_results.get(player1.name, 0) + p1_score
                self.round_robin_results[player2.name] = self.round_robin_results.get(player2.name, 0) + p2_score
        for player in self.strategies:
            self.round_robin_results[player.name] /= self.num_players
        self.strategies.sort(key=lambda x: self.round_robin_results[x.name], reverse=True)

        if show_results:
            print("----------------------------------FINALS---------------------------------------")
        finalist1, finalist2 = self.strategies[:2]
        finals = Arena(finalist1, finalist2, num_rounds=self.num_rounds, show_results=show_results)
        result_dict = finals.play_round().data
        f1_score, f2_score = result_dict[finalist1.name][-1], result_dict[finalist2.name][-1]
        if f1_score > f2_score:
            print(f"The winner is: {finalist1.name}!")
        elif f1_score < f2_score:
            print(f"The winner is: {finalist2.name}")
        else:
            print("Tie!")

    def __play_round_of_16(self, show_results=True):
        results = {
            "Player": [],
            "Reached Quarter Finals": [],
            "Reached Semi Finals": [],
            "Reached Finals": [],
            "Won?": []
        }
        table = {player.name: [False] * 4 for player in self.strategies}
        # ------------------------------------------------------R16--------------------------------------------------
        for i in range(16):
            for j in range(i + 1, 16):
                player1 = self.strategies[i]
                player2 = self.strategies[j]
                arena = Arena(player1, player2, num_rounds=self.num_rounds, show_results=False)
                result_dict = arena.play_round().data
                p1_score, p2_score = result_dict[player1.name][-1], result_dict[player2.name][-1]
                self.round_robin_results[player1.name] = self.round_robin_results.get(player1.name, 0) + p1_score
                self.round_robin_results[player2.name] = self.round_robin_results.get(player2.name, 0) + p2_score
        for player in self.strategies:
            self.round_robin_results[player.name] /= self.num_players
        self.strategies.sort(key=lambda x: self.round_robin_results[x.name], reverse=True)

        for i in range(16):
            if i < 8:
                table[self.strategies[i].name][0] = True
            else:
                table[self.strategies[i].name][0] = False

        # ------------------------------------------------------Quarter Finals----------------------------------------
        for i in range(8):
            for j in range(i + 1, 8):
                player1 = self.strategies[i]
                player2 = self.strategies[j]
                arena = Arena(player1, player2, num_rounds=self.num_rounds, show_results=False)
                result_dict = arena.play_round().data
                p1_score, p2_score = result_dict[player1.name][-1], result_dict[player2.name][-1]
                self.round_robin_results[player1.name] = self.round_robin_results.get(player1.name, 0) + p1_score
                self.round_robin_results[player2.name] = self.round_robin_results.get(player2.name, 0) + p2_score
        for player in self.strategies:
            self.round_robin_results[player.name] /= self.num_players
        self.strategies.sort(key=lambda x: self.round_robin_results[x.name], reverse=True)

        for i in range(8):
            if i < 4:
                table[self.strategies[i].name][1] = True
            else:
                table[self.strategies[i].name][1] = False

        # -------------------------------------------------------Semi Finals------------------------------------------

        for i in range(4):
            for j in range(i + 1, 4):
                player1 = self.strategies[i]
                player2 = self.strategies[j]
                arena = Arena(player1, player2, num_rounds=self.num_rounds, show_results=False)
                result_dict = arena.play_round().data
                p1_score, p2_score = result_dict[player1.name][-1], result_dict[player2.name][-1]
                self.round_robin_results[player1.name] = self.round_robin_results.get(player1.name, 0) + p1_score
                self.round_robin_results[player2.name] = self.round_robin_results.get(player2.name, 0) + p2_score
        for player in self.strategies:
            self.round_robin_results[player.name] /= self.num_players
        self.strategies.sort(key=lambda x: self.round_robin_results[x.name], reverse=True)

        for i in range(4):
            if i < 2:
                table[self.strategies[i].name][2] = True
            else:
                table[self.strategies[i].name][2] = False

        # -------------------------------------------------------Finals------------------------------------------
        if show_results:
            print("----------------------------------FINALS---------------------------------------")

        finalist1, finalist2 = self.strategies[:2]
        finals = Arena(finalist1, finalist2, num_rounds=self.num_rounds, show_results=show_results)
        result_dict = finals.play_round().data
        f1_score, f2_score = result_dict[finalist1.name][-1], result_dict[finalist2.name][-1]
        if f1_score > f2_score:
            print(f"The winner is: {finalist1.name}!")
            table[self.strategies[0].name][3] = True
            table[self.strategies[1].name][3] = False
        elif f1_score < f2_score:
            print(f"The winner is: {finalist2.name}")
            table[self.strategies[1].name][3] = True
            table[self.strategies[0].name][3] = False
        else:
            print("Tie!")
            table[self.strategies[0].name][3] = False
            table[self.strategies[1].name][3] = False

        for player in self.strategies:
            results["Player"].append(player.name)
            results["Reached Quarter Finals"].append(table[player.name][0])
            results["Reached Semi Finals"].append(table[player.name][1])
            results["Reached Finals"].append(table[player.name][2])
            results["Won?"].append(table[player.name][3])

        return Result(results)

    def play(self, show_results):
        self.__reset()
        self.show_info()
        if self.mode == "round robin":
            return self.__play_round_robin(show_results=show_results)
        elif self.mode == "four ways":
            return self.__play_four_ways(show_results=show_results)
        elif self.mode == "round of 16":
            return self.__play_round_of_16(show_results=show_results)




from gametheorylab.axelrod_interactive import get_payoffs
from gametheorylab.axelrod_interactive.strategy import Strategy
import random

class BayesianTitForTat(Strategy):
    """
    Class that implements the Bayesian Tit for Tat strategy. See provided documentation for details.

    Args:
        Note that all parameters apart from 'confidence' must be in (0, 1).
        p_cooperator (float): Perceived probability that opponent mostly cooperates. Defaults to 1/3.
        p_defector (float): Perceived probability that opponent mostly defects. Defaults to 1/3.
        trust (float): Perceived chance that opponent would defect, given that the opponent is believed to be a 'cooperator'. Defaults to 0.1.
        suspicion (float): Perceived chance that opponent would defect, given that the opponent is believed to be a 'defector'. Defaults to 0.9.
        randomness (float): Perceived chance that opponent would defect, given that the opponent is believed to be playing randomly. Defaults to 0.5.
        confidence (int): Number of consistent plays by the opponent before deciding against randomness. Defaults to 5.
        delta_r (float): The extent to which the belief that opponent is playing randomly is reduced, once confidence is reached. Defaults to 0.5.
    """
    payoffs = get_payoffs()
    def __init__(self, p_cooperator=1/3, p_defector=1/3, trust=0.9, suspicion=0.9, randomness=0.5, confidence=5, delta_r=0.5):
        # Validation
        if not 0 < p_cooperator < 1:
            if p_cooperator <= 0:
                string = f"{p_cooperator} <= 0"
            else:
                string = f"{p_cooperator} >= 1"
            raise ValueError(f"Parameter 'p_cooperator' is out of range! ({string})")

        if not 0 < p_defector < 1:
            if p_defector <= 0:
                string = f"{p_defector} <= 0"
            else:
                string = f"{p_defector} >= 1"
            raise ValueError(f"Parameter 'p_defector' is out of range! ({string})")

        if not 0 < trust < 1:
            if trust <= 0:
                string = f"{trust} <= 0"
            else:
                string = f"{trust} >= 1"
            raise ValueError(f"Parameter 'trust' is out of range! ({string})")

        if not 0 < suspicion < 1:
            if suspicion <= 0:
                string = f"{suspicion} <= 0"
            else:
                string = f"{suspicion} >= 1"
            raise ValueError(f"Parameter 'suspicion' is out of range! ({string})")

        if not 0 < randomness < 1:
            if randomness <= 0:
                string = f"{randomness} <= 0"
            else:
                string = f"{randomness} >= 1"
            raise ValueError(f"Parameter 'randomness' is out of range! ({string})")

        if not 0 < delta_r < 1:
            if delta_r <= 0:
                string = f"{delta_r} <= 0"
            else:
                string = f"{delta_r} >= 1"
            raise ValueError(f"Parameter 'delta_r' is out of range! ({string})")

        super().__init__()
        self.name = "Bayesian Tit for Tat"

        self.base_pc = p_cooperator
        self.p_cooperator = p_cooperator

        self.base_pd = p_defector
        self.p_defector = p_defector

        base_pr = 1 - self.p_defector - self.p_cooperator
        self.base_pr = base_pr
        self.p_random = 1 - self.p_defector - self.p_cooperator

        self.base_trust = trust
        self.trust = trust

        self.base_suspicion = suspicion
        self.suspicion = suspicion

        self.base_randomness = randomness
        self.randomness = randomness

        self.confidence = confidence

        self.base_dr = delta_r
        self.delta_r = delta_r
        self.__doc__ = """This strategy uses Bayesian inference to determine the behaviour of the opponent, 
        as determined using the parameters, and responds accordingly.
        
        Using Bayes' Theorem, the strategy's belief of the opponent's cooperation, defection, or randomness is 
        calculated as follows. Denote P(C) by the probability that the opponent is a 'cooperator', i.e. one that 
        mostly cooperates, and similarly define P(D) and P(R). Let P(C|D) be the probability that the opponent 
        cooperates given that the opponent is a defector; define P(D|C), P(C|C), P(D|D), P(C|R), P(D|R) similarly.
        - If the opponent cooperates: 
          Let N = P(C)P(D|C) + P(D)P(D|D) + P(R)P(D|R); then 
          P(C) -> P(C)P(D|C)/N, 
          P(D) -> P(D)P(D|D)/N, and 
          P(R) -> P(R)P(D|R)/N. 
        - If the opponent defects: 
          Let N = P(C)P(C|C) + P(D)P(C|D) + P(R)P(C|R); then 
          P(C) -> P(C)P(C|C)/N, 
          P(D) -> P(D)P(C|D)/N, and 
          P(R) -> P(R)P(C|R)/N."""

    def additional_prep(self):
        self.p_cooperator = self.base_pc
        self.p_defector = self.base_pd
        self.p_random = self.base_pr
        self.trust = self.base_trust
        self.suspicion = self.base_suspicion
        self.randomness = self.base_randomness
        self.delta_r = self.base_dr

    def update_history(self, opp_history, self_history):
        if opp_history:
            confidence_interval = opp_history[-self.confidence:]
            if confidence_interval.count(False) == self.confidence:
                self.randomness *= self.delta_r

            elif 0.4 < confidence_interval.count(False) / self.confidence < 0.6:
                self.randomness /= self.delta_r

            if opp_history[-1]:
                self.p_cooperator *= self.trust
                self.p_defector *= (1 - self.suspicion)
                self.p_random *= self.randomness
            else:
                self.p_cooperator *= (1 - self.trust)
                self.p_defector *= self.suspicion
                self.p_random *= (1 - self.randomness)

            norm = self.p_cooperator + self.p_defector + self.p_random
            self.p_cooperator /= norm
            self.p_defector /= norm
            self.p_random /= norm

    def move(self, opp_history, self_history, opp_score, self_score):
        if not opp_history:
            return True

        return random.random() < self.p_cooperator







from gametheorylab.axelrod_interactive.strategy import Strategy

class GrimTrigger(Strategy):
    """Class that implements the Grim Trigger strategy."""
    def __init__(self):
        super().__init__()
        self.name = "Grim Trigger"
        self.__doc__ = ("This strategy cooperates continually, until the opponent defects once, and then defects for "
                        "the remainder of the game.")

    def move(self, opp_history, self_history, opp_score, self_score):
        return self.opps_defect_count == 0



from gametheorylab.axelrod_interactive.strategy import Strategy

class User(Strategy):
    """Class that implements the User strategy - an interface through which the user can play the game!"""
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.__doc__ = (f"This is a human, going by the name of {self.name}. No one knows how this player's going to "
                        f"dole out his moves. Let's find out, shall we?")

    def move(self, opp_history, self_history, opp_score, self_score):
        print("Your turn!")

        print("Here what your opponent has played so far: ")
        print(["C" if move else "D" for move in opp_history])

        print("And here's what YOU've played so far!")
        print(["C" if move else "D" for move in self_history])

        print(f"Your current score: {self_score}")
        print(f"Your opponent's score: {opp_score}")

        print(f"Over to you, {self.name}. Let's see if you can outdo your opponent!")
        choice = input("Choose a move. Type 'C' for 'Cooperate' or 'D' for 'Defect': ").title()
        while choice not in {"C", "D"}:
            print("I said type 'C' for 'Cooperate' or 'D' for 'Defect'! C'mon, follow instructions :( ")
            choice = input("Choose a move. Type 'C' for 'Cooperate' or 'D' for 'Defect': ").title()

        return choice == "C"

